/*
 * Copyright (c) 1997, 1998
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id$
 */

/*
 * This module implements a probe/attach mechanism for the MII layer.
 * This code is adapted from NetBSD, using a couple of ugly hacks
 * necessitated by FreeBSD's lack of support for the device autoconfig
 * magic in /sys/lern/subr_autoconf.c that NetBSD uses. Personally,
 * I think it's too bad that FreeBSD doesn't support this stuff, but
 * I still hold out hope that someday somebody will either implement
 * it all or come up with an equally useful alternative. In the meantime,
 * we need to kludge up a few things.
 *
 * In NetBSD, each device has 'struct device' associated with it, which it
 * passes down to the mii probe routines. These in turn attach a child
 * device to the corresponding parent (in this case, the parent is the
 * ethernet controller and the child is the PHY attached to it). We fake
 * this behavior by using the ifnet structure instead. We can then use
 * this as an argument for the callbacks, plus it contains the interface
 * name of the parent device.
 *
 * The second issue is the autoconfiguration. We don't want to have an
 * miidevsw[] struct with hardcoded entries for each PHY driver since that
 * would make it harder to add new ones later. Instead, we use a linker
 * set in the same fashion as PCI drivers. This module contains a search
 * routine to query all the PHY drivers in search of one that handles the
 * given PHY device, with the 'unknown' driver used in case no exact
 * matches are found. This takes the place of the config_found_sm() routine,
 * which we don't have. Among other things, this saves us from having to
 * specify each PHY driver in the kernel config file.
 */

#include <sys/param.h>
#include <sys/device.h>
#include <sys/systm.h>
#include <sys/socket.h>
#include <sys/malloc.h>

#include <net/if.h>
#include <net/if_media.h>

#include <dev/mii/mii.h>
#include <dev/mii/miivar.h>
#include <dev/mii/mii_glue.h>

#ifndef lint
static char rcsid[] =
	"$Id$";
#endif

struct mii_softc *mii_match(parent, ma, mii_print, mii_unit)
	struct ifnet		*parent;
	struct mii_attach_args	*ma;
	int (*mii_print)(void *, const char *);
{
	struct mii_cfattach	**mpp;
	struct mii_cfattach	*mp = NULL, *miidev = NULL;
	int			matchprio = 0, tmpprio = 0;
	struct mii_softc	*sc;
	char			ifname[16];

	mpp = (struct mii_cfattach **)miidevice_set.ls_items;

	while((mp = *mpp++) != NULL) {
		tmpprio = (*mp->mii_cfmatch)(parent, NULL, ma);
		if (tmpprio > matchprio) {
			matchprio = tmpprio;
			miidev = mp;
		}
	}

	if (miidev == NULL)
		return(NULL);

	sc = malloc(miidev->mii_softcsize, M_DEVBUF, M_NOWAIT);
	if (sc == NULL)
		return(NULL);
	bzero((char *)sc, miidev->mii_softcsize);

	sc->mii_dev.dv_unit = *miidev->mii_count;
	*miidev->mii_count++;
	sprintf(sc->mii_dev.dv_xname, "%s%d", miidev->mii_name,
					sc->mii_dev.dv_unit);
	sc->mii_dev.dv_parent = (struct device *)parent;
	printf("mii%d: ", mii_unit);
	sprintf(ifname, "%s%d", parent->if_name, parent->if_unit);
	mii_print(ma, ifname);
	printf("\n");
	printf("%s", sc->mii_dev.dv_xname);
	(*miidev->mii_cfattach)(parent, sc, ma);

	return(sc);
}
